# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#GLOBAL CONSTANTS are defined here
TRAIN_DATA_POINT_MARKER = -1
TEST_DATA_POINT_MARKER = -2

TRAIN_2D_FILE_NAME = "train_2d.npy"
TEST_2D_FILE_NAME = "test_2d.npy"

INVERSE_PROJECTION_ERRORS_FILE = "inverse_projection_errors.npy"
PROJECTION_ERRORS_INTERPOLATED_FILE = "projection_errors_interpolated.npy"
PROJECTION_ERRORS_INVERSE_PROJECTION_FILE = "projection_errors_inv_projection.npy"


#GUI CONSTANTS
APP_FONT = "Helvetica 12"
APP_FONT_BOLD = "Helvetica 12 bold"
BLACK_COLOR = "#252526"
APP_PRIMARY_COLOR = "#007acc"
WHITE_COLOR = "#ffffff"
RED_COLOR = "#ff0000"
GREEN_COLOR = "#00ff00"
YELLOW_COLOR = "#ffff00"
RIGHTS_MESSAGE_1 = "© 2023 Cristian Grosu. All rights reserved."
RIGHTS_MESSAGE_2 = "Made by Cristian Grosu for Utrecht University Master Thesis in 2023"